using System;
using System.Collections.Generic;
using System.Text;
using System.IO; // DirecoryInfo
using System.Diagnostics; // ProcessInfo
using System.Collections.Specialized; // StringCollection


namespace CheckInclude
{
    class CheckInclude
    {
        //-----  Attributes  -----

        private Log      m_log;
        private Settings m_settings;

        private string m_cppCheckFileFullName;
        private string m_cppCheckFileBackupFullName;
        private string m_backupFileFullName;

        private DateTime m_startTime;
        private int m_numErrorsCompileIncludes = 0;
        private int m_numErrorsCompileCppFiles = 0;
        


        //-----  Public methods  -----

        static void Main(string[] args)
        {
            try
            {
                CheckInclude checker = new CheckInclude(args);
                checker.Run();
            }
            catch (SystemException ex)
            {
                Console.WriteLine();
                Console.WriteLine("EXCEPTION: " + ex.Message);
            }

        }

        //-----------------------------------------------------------

        // Constructor
        public CheckInclude(
            string[] in_args)
        {
            try
            {
                // Create log file
                m_log = new Log();

                // Read settings from command line
                m_settings = new Settings(in_args, m_log);
            }
            catch (SystemException ex)
            {
                m_log.WriteTo(Log.ALL, "");
                m_log.WriteTo(Log.ALL, "EXCEPTION: " + ex.Message);
                throw new SystemException("Check was cancelled (" + ex.Message + ")");
            }
        }


        //-----------------------------------------------------------


        public void Run()
        {
            try
            {
                m_startTime = DateTime.Now;
                m_log.WriteTo(Log.ALL, "________________________________________________________________________________");
                m_log.WriteTo(Log.ALL, "\n" + GetCurDateTime(m_startTime) + " Starting check");
                m_log.WriteTo(Log.ALL, "________________________________________________________________________________");

                m_log.WriteTo(Log.ALL, "\r\nSTEP 1: Copying search dir to temporary folder " + m_settings.NameSearchDirTemp);
                CopySearchDirToTemporaryFolder(m_settings.NameSearchDir, m_settings.NameSearchDirTemp);

                m_log.WriteTo(Log.ALL, "\r\nSTEP 2: Verify existence of cpp file to compile includes");
                CheckForSpecialCppFile();

                m_log.WriteTo(Log.ALL, "\r\nSTEP 3: Verify that project compiles");
                CheckProjectCompiles();

                m_log.WriteTo(Log.ALL, "\r\nSTEP 4: Verify that all includes compile alone");
                CheckAllIncludesCompileAlone();

                m_log.WriteTo(Log.ALL, "\r\nSTEP 5: Check all CPP files for unneeded includes");
                CheckCppFilesForUnneededIncludes();

                DateTime stopTime = DateTime.Now;
                TimeSpan duration = stopTime - m_startTime;
                m_log.WriteTo(Log.ALL, "________________________________________________________________________________");
                m_log.WriteTo(Log.ALL, "\n" + GetCurDateTime(stopTime) + " Check finished, duration: " + duration.ToString());
                m_log.WriteTo(Log.ALL, "\nCheck Includes  : " + m_numErrorsCompileIncludes + " errors detected");
                m_log.WriteTo(Log.ALL, "Check cpp files : " + m_numErrorsCompileCppFiles + " errors detected");
                m_log.WriteTo(Log.ALL, "________________________________________________________________________________");

            }
            catch (SystemException ex)
            {
                m_log.WriteTo(Log.ALL, "");
                m_log.WriteTo(Log.ALL, "EXCEPTION: " + ex.Message);
            }
        }

        //-----------------------------------------------------------


        private void CheckProjectCompiles()
        {
            string appName  = "devenv.exe";
            string buildCmd = "";

            if (m_settings.CmdToCompileProject == Settings.BuildType.SKIP)
            {
                m_log.WriteTo(Log.ALL, "INFO: Test to check compilability of project is skipped");
                return;
            }
            else
            {
                buildCmd = (m_settings.CmdToCompileProject == Settings.BuildType.BUILD) ?
                    "/Build" : "/Rebuild";
            }
            string cmdLineArgsCompiler = m_settings.NameProjectFile + " " + buildCmd + " Debug /Out "
                + m_log.m_nameLogFileFullPath;
            int exitCode = 7;
            m_log.WriteTo(Log.LOG, "\nVerify that project is compilable:");
            m_log.WriteTo(Log.LOG, "Calling " + appName + " " + cmdLineArgsCompiler);
            m_log.WriteTo(Log.LOG, GetCurDateTime() + " Starting build...");
            exitCode = StartApplication(appName, cmdLineArgsCompiler, false, true);
            m_log.WriteTo(Log.LOG, GetCurDateTime() + " ... build terminated with exit code " + exitCode);

            if (exitCode!=0)
            {
                throw new SystemException("FATAL ERROR: Building project returns with exit code " + exitCode);
            }
            else
            {
                m_log.WriteTo(Log.ALL, "OK Building project was successful");
            }
        }

        //-----------------------------------------------------------

        private void CheckForSpecialCppFile()
        {
            // Project must have a cpp file as given within command line (default is "ToCheckInclude.cpp")
            // Rule: include file is appended at end of this (possibly empty) file (e.g. preceding stdafx.h remains unchanged)

            FileInfo cppCheckFile = new FileInfo(m_settings.PathCppFileToCompileIncludes);
            if (!cppCheckFile.Exists)
            {
                throw new SystemException("FATAL ERROR: required cpp file for checking includes not found: " + m_settings.PathCppFileToCompileIncludes);
            }
            m_cppCheckFileFullName = cppCheckFile.FullName;
            m_cppCheckFileBackupFullName = cppCheckFile.FullName + ".Backup";
            m_backupFileFullName = m_settings.NameSearchDirTemp + "/TEMP_FILE.Backup";
            m_log.WriteTo(Log.ALL, "OK found cppFileToCompileIncludes at : " + m_cppCheckFileFullName);
            m_log.WriteTo(Log.LOG, "Using cpp backup file                : " + m_cppCheckFileBackupFullName);
            m_log.WriteTo(Log.LOG, "Using header backup file             : " + m_backupFileFullName);
            File.Copy(m_cppCheckFileFullName, m_cppCheckFileBackupFullName, true);
        }

        //-----------------------------------------------------------


        private void CheckAllIncludesCompileAlone()
        {
            System.Collections.Specialized.StringCollection foldersSkipped = new StringCollection();
            System.Collections.Specialized.StringCollection filesSkipped = new StringCollection();
            System.Collections.Specialized.StringCollection headerFiles = FindFiles(m_settings.NameSearchDirTemp,
            "*.h", true, null, m_settings.DirsToSkip, foldersSkipped, filesSkipped);

            m_log.WriteTo(Log.LOG, "\nFound " + headerFiles.Count + " header files to check:");
            for (int i = 0; i < headerFiles.Count; ++i)
            {
                m_log.WriteTo(Log.LOG, headerFiles[i]);
            }

            if (foldersSkipped.Count > 0)
            {
                m_log.WriteTo(Log.LOG, "\nThe following folders were skipped:");
                for (int i = 0; i < foldersSkipped.Count; ++i)
                {
                    m_log.WriteTo(Log.LOG, foldersSkipped[i]);
                }
            }

            if (filesSkipped.Count > 0)
            {
                m_log.WriteTo(Log.LOG, "\nThe following files were skipped:");
                for (int i = 0; i < filesSkipped.Count; ++i)
                {
                    m_log.WriteTo(Log.LOG, filesSkipped[i]);
                }
            }

            if (m_settings.CmdToCompileInclude == Settings.BuildType.SKIP)
            {
                m_log.WriteTo(Log.ALL, "INFO: Test to check compilability of includes is skipped");
                return;
            }

            int exitCode;
            for (int includeNum = 1; includeNum <= headerFiles.Count; ++includeNum)
            {
                m_log.WriteTo(Log.STDOUT_LOG, "\r\nChecking include file " + includeNum + " of " + headerFiles.Count + " : " + headerFiles[includeNum - 1]);

                // Add include of found include file to cpp file used for compiling
                File.Copy(m_cppCheckFileBackupFullName, m_cppCheckFileFullName, true);
                StreamWriter cppWriter = new StreamWriter(m_cppCheckFileFullName,
                    true, Encoding.GetEncoding("windows-1252"));
                string incFileFullPath = headerFiles[includeNum - 1];
                incFileFullPath = incFileFullPath.Replace("\\", "/");
                cppWriter.WriteLine("#include \"" + incFileFullPath + "\"");
                cppWriter.Close();

                // Build command to call compiler
                string appName = "devenv.exe";
                string buildCmd = (m_settings.CmdToCompileInclude == Settings.BuildType.BUILD) ?
                        "/Build" : "/Rebuild";
                string cmdLineArgsCompiler = m_settings.NameProjectFile + " " + buildCmd + " Debug /Out "
                    + m_log.m_nameLogFileFullPath;

                m_log.WriteTo(Log.LOG, "Calling " + appName + " " + cmdLineArgsCompiler);

                m_log.WriteTo(Log.LOG, GetCurDateTime() + " Starting build...");
                exitCode = StartApplication(appName, cmdLineArgsCompiler, false, true);
                m_log.WriteTo(Log.LOG, GetCurDateTime() + " ... build terminated with exit code " + exitCode);
                if (exitCode == 0)
                {
                    m_log.WriteTo(Log.LOG, "OK Include compiles " + incFileFullPath);
                }
                else
                {
                    ++m_numErrorsCompileIncludes;
                    m_log.WriteTo(Log.ALL, "ERROR Include does not compile " + incFileFullPath);
                }

                if ((exitCode == 0) && m_settings.CheckForUnneededIncludes)
                {
                    GenericCheckForUnneededIncludes(includeNum, headerFiles.Count, headerFiles[includeNum - 1],
                        m_settings.CmdToCompileInclude, ref m_numErrorsCompileIncludes);
                }
            }

            if (m_numErrorsCompileIncludes == 0)
            {
                m_log.WriteTo(Log.ALL,"OK all checks of include files succeeded");
            }
            else
            {
                m_log.WriteTo(Log.ALL,"ERROR: found " + m_numErrorsCompileIncludes + " problems during check of include files");
            }

            // Restore original cpp file
            File.Copy(m_cppCheckFileBackupFullName, m_cppCheckFileFullName, true);
        }

        //-----------------------------------------------------------

        private void CheckCppFilesForUnneededIncludes()
        {
            System.Collections.Specialized.StringCollection foldersSkipped = new StringCollection();
            System.Collections.Specialized.StringCollection filesSkipped = new StringCollection();
            System.Collections.Specialized.StringCollection cppFiles = FindFiles(m_settings.NameSearchDirTemp,
            "*.cpp", true, null, m_settings.DirsToSkip, foldersSkipped, filesSkipped);

            m_log.WriteTo(Log.LOG, "\nFound " + cppFiles.Count + " CPP files to check:");
            for (int i = 0; i < cppFiles.Count; ++i)
            {
                m_log.WriteTo(Log.LOG, cppFiles[i]);
            }

            if (foldersSkipped.Count > 0)
            {
                m_log.WriteTo(Log.LOG, "\nThe following folders were skipped:");
                for (int i = 0; i < foldersSkipped.Count; ++i)
                {
                    m_log.WriteTo(Log.LOG, foldersSkipped[i]);
                }
            }

            if (filesSkipped.Count > 0)
            {
                m_log.WriteTo(Log.LOG, "\nThe following files were skipped:");
                for (int i = 0; i < filesSkipped.Count; ++i)
                {
                    m_log.WriteTo(Log.LOG, filesSkipped[i]);
                }
            }

            if (m_settings.CmdToCompileCpp == Settings.BuildType.SKIP  || !m_settings.CheckForUnneededIncludes)
            {
                m_log.WriteTo(Log.ALL, "INFO: Test to check cpp files for unneeded includes is skipped");
                return;
            }

            m_numErrorsCompileCppFiles = 0;
            for (int cppFileNum = 1; cppFileNum <= cppFiles.Count; ++cppFileNum)
            {
                m_log.WriteTo(Log.STDOUT_LOG, "\r\nChecking cpp file " + cppFileNum + " of " + cppFiles.Count + " : " + cppFiles[cppFileNum - 1]);
                GenericCheckForUnneededIncludes(cppFileNum, cppFiles.Count, cppFiles[cppFileNum - 1],
                    m_settings.CmdToCompileCpp, ref m_numErrorsCompileCppFiles);
            }

            if (m_numErrorsCompileCppFiles == 0)
            {
                m_log.WriteTo(Log.ALL, "OK all checks of cpp files succeeded");
            }
            else
            {
                m_log.WriteTo(Log.ALL, "ERROR: found " + m_numErrorsCompileCppFiles + " problems during check of cpp files");
            }
        }

        //-----------------------------------------------------------

        private void GenericCheckForUnneededIncludes(
            int in_numCurFile,
            int in_numAllFiles,
            string in_nameFile,
            Settings.BuildType in_compileType,
            ref int io_numErrors)
        {
            m_log.WriteTo(Log.LOG, "CheckForUnneededIncludes " + in_nameFile);

            // Save original file
            File.Copy(in_nameFile, m_backupFileFullName, true);

            bool includeRemoved = false;
            int lastLineNumWithRemovedInclude = 0;
            string lastRemovedInclude = "";
            int curNumIncludeStatement = 0;
            System.Collections.SortedList removedIncludes = new System.Collections.SortedList();

            do
            {
                // Try to find the next include statement within file and
                // remove it to check if file still compiles
                StreamWriter writer = new StreamWriter(in_nameFile,
                    false, Encoding.GetEncoding("windows-1252"));

                StreamReader reader = new StreamReader(m_backupFileFullName, Encoding.GetEncoding("windows-1252"));

                includeRemoved = false;
                int curLineNum = 0;
                string line;
                bool countNumIncludeStatements = true;
                int numIncludeStatements = 0;
                while ((line = reader.ReadLine()) != null)
                {
                    ++curLineNum;
                    if (line.StartsWith("#include"))
                    {
                        if (countNumIncludeStatements)
                        {
                            ++numIncludeStatements;
                        }

                        // all includes which have been found as not necessary will remain removed
                        if (removedIncludes.Contains(curLineNum))
                        {
                            continue;
                        }

                        if ((curLineNum > lastLineNumWithRemovedInclude) && (!includeRemoved))
                        {
                            includeRemoved = true;
                            lastLineNumWithRemovedInclude = curLineNum;
                            lastRemovedInclude = line;
                        }
                        else // simply copy line
                        {
                            writer.WriteLine(line);
                        }
                    }
                    else
                    {
                        writer.WriteLine(line);
                    }
                }
                countNumIncludeStatements = false;

                writer.Close();
                reader.Close();

                if (includeRemoved)
                {
                    ++curNumIncludeStatement;

                    m_log.WriteTo(Log.STDOUT_LOG,
                        in_numCurFile + " of " + in_numAllFiles + " / " +
                        curNumIncludeStatement + " of " + numIncludeStatements +
                        " Testing include statement at line " + lastLineNumWithRemovedInclude + " (" + lastRemovedInclude + ")");

                    // Call compiler
                    string appName = "devenv.exe";
                    string buildCmd = (in_compileType == Settings.BuildType.BUILD) ?
                            "/Build" : "/Rebuild";
                    string cmdLineArgsCompiler = m_settings.NameProjectFile + " " + buildCmd + " Debug /Out "
                        + m_log.m_nameLogFileFullPath;
                    m_log.WriteTo(Log.LOG,
                        in_numCurFile + " of " + in_numAllFiles + " / " +
                        curNumIncludeStatement + " of " + numIncludeStatements + " Calling " + appName + " " + cmdLineArgsCompiler);
                    m_log.WriteTo(Log.LOG, GetCurDateTime() + " Starting build...");
                    int exitCode = StartApplication(appName, cmdLineArgsCompiler, false, true);
                    m_log.WriteTo(Log.LOG, GetCurDateTime() + " ... build terminated with exit code " + exitCode);

                    if (exitCode == 0)
                    {
                        ++io_numErrors;
                        m_log.WriteTo(Log.ALL, "  ERROR: include statement is not necessary, line " + lastLineNumWithRemovedInclude
                            + " (" + lastRemovedInclude + ") in file " + in_nameFile);
                        // Store as unneeded include for further checks
                        removedIncludes.Add(lastLineNumWithRemovedInclude, 1);
                    }
                    else
                    {
                        m_log.WriteTo(Log.LOG, "OK include is needed, line " + lastLineNumWithRemovedInclude
                            + " (" + lastRemovedInclude + ")");
                    }
                }

            } while (includeRemoved);


            // Restore original cpp file
            File.Copy(m_backupFileFullName, in_nameFile, true);
        }


        //-----------------------------------------------------------

        // Delegate to indicate progress
        public delegate void FindProgress(string folderName);

        // Method to search files within a folder
        private System.Collections.Specialized.StringCollection FindFiles(
            string folderName,
            string searchPattern,
            bool recurse,
            FindProgress findProgress,
            System.Collections.Specialized.StringCollection folderNamesToSkip,
            System.Collections.Specialized.StringCollection foldersSkipped,
            System.Collections.Specialized.StringCollection filesSkipped)
        {
            // StringCollection-Objekt fr die Rckgabe erzeugen
            StringCollection resultFiles = new StringCollection();

            // DirectoryInfo-Objekt fr den Ordner erzeugen
            DirectoryInfo folder = new DirectoryInfo(folderName);

            // Die rekursive Methode zum Suchen in einem Ordner aufrufen
            FindFilesInFolder(searchPattern, folder, recurse, resultFiles,
               findProgress, folderNamesToSkip, foldersSkipped, filesSkipped);

            // StringCollection zurckgeben
            return resultFiles;
        }

        //-----------------------------------------------------------

        // recursive method to look for files
        private void FindFilesInFolder(
            string                       searchPattern,
            System.IO.DirectoryInfo folder,
            bool recurse,
            System.Collections.Specialized.StringCollection resultFiles,
            FindProgress findProgress,
            System.Collections.Specialized.StringCollection folderNamesToSkip,
            System.Collections.Specialized.StringCollection foldersSkipped,
            System.Collections.Specialized.StringCollection filesSkipped)
        {
            // call delegate if existing
            if (findProgress != null)
                findProgress(folder.FullName);

            // if required call recursive for all subfolders
            if (recurse)
            {
                DirectoryInfo[] subFolders = folder.GetDirectories();
                for (int i = 0; i < subFolders.Length; i++)
                {
                    if (folderNamesToSkip.Contains(subFolders[i].Name))
                    {
                        foldersSkipped.Add(subFolders[i].FullName);
                    }
                    else
                    {
                        FindFilesInFolder(searchPattern, subFolders[i], true, resultFiles,
                           findProgress, folderNamesToSkip, foldersSkipped, filesSkipped);
                    }
                }
            }

            // Find files mathing with search pattern
            FileInfo[] files = folder.GetFiles(searchPattern);

            // Check found files. If not to skip store full path name
            // within StringCollection
            for (int i = 0; i < files.Length; i++)
            {
                if (m_settings.FilesToSkip.Contains(files[i].Name))
                {
                    filesSkipped.Add(files[i].FullName);
                }
                else
                {
                    resultFiles.Add(files[i].FullName);
                }
            }
        }

        //-----------------------------------------------------------

        // method to start an application, optional wait for termination
        public static int StartApplication(string exeFileName, string arguments,
           bool maximized, bool waitForExit)
        {
            int retExitCode;

            // generate and init ProcessStartInfo instance
            ProcessStartInfo psi = new ProcessStartInfo(exeFileName);
            if (arguments != "")
                psi.Arguments = arguments;
            if (maximized)
                psi.WindowStyle = ProcessWindowStyle.Maximized;
            psi.UseShellExecute = true;

            // start process
            Process process = Process.Start(psi);

            // Wait for process termination if required
            if (waitForExit)
                process.WaitForExit();

            retExitCode = process.ExitCode;
            return retExitCode;
        }


        //-----------------------------------------------------------

        // method to copy the search directory to a (n ot existing) temporary folder
        public void CopySearchDirToTemporaryFolder(string sourceFolderName, string destFolderName)
        {
            // Create DirectoryInfo object for source folder
            DirectoryInfo sourceFolder = new DirectoryInfo(sourceFolderName);

            // Check if target folder already exists
            if (Directory.Exists(destFolderName))
            {
                if (m_settings.ForceCopy)
                {
                    throw new IOException("Der Zielordner '" + destFolderName +
                       "' existiert bereits");
                }
                else
                {
                    m_log.WriteTo(Log.ALL, "INFO: Target folder '" + destFolderName + "' already exists. Assuming files are up to date and copy can be skipped!");
                    return;
                }
            }

            // Create target folder
            Directory.CreateDirectory(destFolderName);

            // Copy all files and subfolders
            CopySubFoldersAndFiles(sourceFolder, sourceFolderName, destFolderName);

            m_log.WriteTo(Log.ALL, "OK files and folders successfully copied");
        }

        //-----------------------------------------------------------

        // helper method to recursively copy a folder
        private static void CopySubFoldersAndFiles(
            DirectoryInfo folder,
            string sourceFolderName,
            string destFolderName)
        {
            // Iterate over subfolders
            DirectoryInfo[] subFolders = folder.GetDirectories();
            for (int i = 0; i < subFolders.Length; i++)
            {
                // Build name for target folder by replacing path to parent source folder
                // with path to parent target folder
                string destSubFolderName = subFolders[i].FullName.Replace(
                   sourceFolderName, destFolderName);

                // Create subfolder within target folder
                Directory.CreateDirectory(destSubFolderName);

                // Call method recursively to first create all remainiung subfolders
                CopySubFoldersAndFiles(subFolders[i], sourceFolderName,
                   destFolderName);
            }

            // Get al contained files
            FileInfo[] files = folder.GetFiles();

            // Iterate over all files
            for (int i = 0; i < files.Length; i++)
            {
                // Build name for target file by replacing path to parent source folder
                // with path to parent target folder
                string destFileName = files[i].FullName.Replace(
                   sourceFolderName, destFolderName);

                // Copy file
                File.Copy(files[i].FullName, destFileName);
            }
        }

        //-----------------------------------------------------------

        private string GetCurDateTime()
        {
            DateTime now = DateTime.Now;
            return GetCurDateTime(now);
        }

        private string GetCurDateTime(DateTime in_time)
        {
            string result = in_time.ToShortDateString() + " " + in_time.ToLongTimeString();
            return result;
        }

    }
}
