using System;
using System.Collections.Generic;
using System.Text;
using System.Collections.Specialized; // StringCollection
using System.IO; // DirecoryInfo


namespace CheckInclude
{
    class Settings
    {
        //-----------------------------------------------------------
        // Attributes
        //-----------------------------------------------------------

        public enum BuildType { REBUILD, BUILD, SKIP };

        // Build command to check several aspects
        private BuildType m_cmdToCheckIfProjectCompiles;
        private BuildType m_cmdToCheckIfIncludeCompiles;
        private BuildType m_cmdToCheckIfCppCompiles;

        // Project file used for compiling
        // Remark: Currently it is assumed that project file path is
        // base path (arg[0]) + arg[1]
        private string m_nameProjectFile;

        // Search directory, within this directory and all its subdirs
        // all includes are checked
        // Remark: Currently it is assumed thatseatrch dir is
        // base path (arg[0]) + arg[2] (which may be ".")
        private string m_nameSearchDir;

        // Temporary search dir used for compiling. The original sources are copied to
        // this destination
        private string m_nameSearchDirTemp;

        // CPP file used for dynamically inserting include names
        // and checking for compilability
        private string m_relativePathCppFileToCompileIncludes;
        private string m_pathCppFileToCompileIncludes;

        // Directorys which are not checked (e.g. may contain automatically generated code)
        private System.Collections.Specialized.StringCollection m_dirsToSkip;

        // Files which are not checked (e.g. stdafx.h, stdafx.cpp)
        private System.Collections.Specialized.StringCollection m_filesToSkip;

        // Flag indicating whether the source directory has to be copied completely
        // If false the source directory is only copied if the temporary directory
        // does not exist.
        private bool m_forceCopy;

        // Flag indicating if a check for unnecessary include statements ius performed.
        private bool m_checkForUnneededIncludes;

        // Command line arguments
        private string[] m_args;

        // Access to log files
        private Log m_log;


        //-----------------------------------------------------------
        // Properties
        //-----------------------------------------------------------

        public BuildType CmdToCompileProject
        {
            get { return m_cmdToCheckIfProjectCompiles; }
        }

        public BuildType CmdToCompileInclude
        {
            get { return m_cmdToCheckIfIncludeCompiles; }
        }

        public BuildType CmdToCompileCpp
        {
            get { return m_cmdToCheckIfCppCompiles; }
        }

        public string NameProjectFile
        {
            get { return m_nameProjectFile; }
        }

        public string NameSearchDir
        {
            get { return m_nameSearchDir; }
        }

        public string NameSearchDirTemp
        {
            get { return m_nameSearchDirTemp; }
        }

        public string PathCppFileToCompileIncludes
        {
            get { return m_pathCppFileToCompileIncludes; }
        }

        public System.Collections.Specialized.StringCollection DirsToSkip
        {
            get { return m_dirsToSkip; }
        }

        public System.Collections.Specialized.StringCollection FilesToSkip
        {
            get { return m_filesToSkip; }
        }

        public bool ForceCopy
        {
            get { return m_forceCopy; }
        }

        public bool CheckForUnneededIncludes
        {
            get { return m_checkForUnneededIncludes; }
        }

        //-----------------------------------------------------------
        // Public methods
        //-----------------------------------------------------------

        // Constructor
        public Settings(string[] in_args, Log in_log)
        {
            m_args = in_args;
            m_log  = in_log;

            // Set default values
            m_relativePathCppFileToCompileIncludes = "ToCheckIncludes.cpp";
            m_forceCopy = true;
            m_checkForUnneededIncludes = true;
            m_cmdToCheckIfProjectCompiles = BuildType.BUILD;
            m_cmdToCheckIfIncludeCompiles = BuildType.BUILD;
            m_cmdToCheckIfCppCompiles     = BuildType.BUILD;
            m_dirsToSkip = new StringCollection();
            m_filesToSkip = new StringCollection();

            // Write name of application and executable path to log file
            System.IO.FileInfo fi = new System.IO.FileInfo(System.Reflection.Assembly.GetEntryAssembly().Location);
            string appDirName = fi.DirectoryName;
            string appName = fi.Name;
            m_log.WriteTo(Log.ALL, "Executable name           : " + appName);
            m_log.WriteTo(Log.ALL, "Executable directory      : " + appDirName);

            ReadCommandLine();
            WriteSettingsToLog();
        }


        //-----------------------------------------------------------
        // Private methods
        //-----------------------------------------------------------
        
        private void ReadCommandLine()
        {
            m_log.WriteTo(Log.LOG, "Number of arguments       : " + m_args.Length);
            for (int i = 0; i < m_args.Length; ++i)
            {
                m_log.WriteTo(Log.LOG, "arg " + i + "                     : " + m_args[i]);
            }
            if (m_args.Length < 3)
            {
                string info = "Too few arguments! Syntax: CheckInclude  <absolute base path> <relative path to project file>" +
                    " <relative path to search dir>";
                m_log.WriteTo(Log.ALL, info);
                throw new SystemException(info);
            }

            string basePath = m_args[0];
            m_nameSearchDir = Path.Combine(basePath, m_args[2]);

            DirectoryInfo sourceDirInfo = new DirectoryInfo(m_nameSearchDir);
            m_nameSearchDirTemp = Path.Combine(sourceDirInfo.Parent.FullName, "TEMP_CHECK_INCLUDE_" + sourceDirInfo.Name);

            m_nameProjectFile = Path.Combine(m_nameSearchDirTemp, m_args[1]);

            int curArgIdx=0;
            try
            {
                // Analyze specific arguments
                for (int i = 3; i < m_args.Length; ++i)
                {
                    curArgIdx = i;
                    string paramValue;
                    if (ReadParam("skipDir", m_args[i], out paramValue))
                    {
                        m_dirsToSkip.Add(paramValue);
                    }
                    else if (ReadParam("skipFile", m_args[i], out paramValue))
                    {
                        m_filesToSkip.Add(paramValue);
                    }
                    else if (ReadParam("forceCopy", m_args[i], out paramValue))
                    {
                        m_forceCopy = (paramValue == "YES");
                    }
                    else if (ReadParam("checkForUnneededIncludes", m_args[i], out paramValue))
                    {
                        m_checkForUnneededIncludes = (paramValue == "YES");
                    }
                    else if (ReadParam("cmdToCompileProject", m_args[i], out paramValue))
                    {
                        m_cmdToCheckIfProjectCompiles = (Settings.BuildType)Enum.Parse(typeof(Settings.BuildType), paramValue);
                    }
                    else if (ReadParam("cmdToCompileInclude", m_args[i], out paramValue))
                    {
                        m_cmdToCheckIfIncludeCompiles = (Settings.BuildType)Enum.Parse(typeof(Settings.BuildType), paramValue);
                    }
                    else if (ReadParam("cmdToCompileCppFile", m_args[i], out paramValue))
                    {
                        m_cmdToCheckIfCppCompiles = (Settings.BuildType)Enum.Parse(typeof(Settings.BuildType), paramValue);
                    }
                    else if (ReadParam("cppFileToCompileIncludes", m_args[i], out paramValue))
                    {
                        m_relativePathCppFileToCompileIncludes = paramValue;
                    }
                }
                m_pathCppFileToCompileIncludes = Path.Combine(m_nameSearchDirTemp, m_relativePathCppFileToCompileIncludes);
            }
            catch (System.Exception e)
            {
                throw new SystemException("FATAL ERROR when parsing arg num " + curArgIdx + ", '" + m_args[curArgIdx] + "' (" + e.Message + ")");
            }
        }

        //-----------------------------------------------------------

        private void WriteSettingsToLog()
        {
            m_log.WriteTo(Log.ALL, "Logfile                   : " + m_log.m_nameLogFileFullPath);
            m_log.WriteTo(Log.ALL, "ResultFile                : " + m_log.m_nameResultFileFullPath);
            m_log.WriteTo(Log.ALL, "SearchDir                 : " + m_nameSearchDir);
            m_log.WriteTo(Log.ALL, "Temp SearchDir            : " + m_nameSearchDirTemp);
            m_log.WriteTo(Log.ALL, "ProjectFile               : " + m_nameProjectFile);
            m_log.WriteTo(Log.ALL, "forceCopy                 : " + m_forceCopy);
            m_log.WriteTo(Log.ALL, "checkForUnneededIncludes  : " + m_checkForUnneededIncludes);
            m_log.WriteTo(Log.ALL, "cmdToCompileProject       : " + m_cmdToCheckIfProjectCompiles);
            m_log.WriteTo(Log.ALL, "cmdToCompileInclude       : " + m_cmdToCheckIfIncludeCompiles);
            m_log.WriteTo(Log.ALL, "cmdToCompileCpp           : " + m_cmdToCheckIfCppCompiles);
            m_log.WriteTo(Log.ALL, "cppFileToCompileIncludes  : " + m_pathCppFileToCompileIncludes);

            string dirsToSkip = "";
            foreach (string dir in m_dirsToSkip)
            {
                dirsToSkip += " " + dir;
            }
            m_log.WriteTo(Log.ALL, "dirs to skip              :" + dirsToSkip);

            string filesToSkip = "";
            foreach (string file in m_filesToSkip)
            {
                filesToSkip += " " + file;
            }
            m_log.WriteTo(Log.ALL, "files to skip             :" + filesToSkip);
        }

        //-----------------------------------------------------------
        
        private bool ReadParam(string in_paramName, string in_argStr, out string out_paramValue)
        {
            string completeParamExpression = "/" + in_paramName + "=";
            out_paramValue = "";
            if (in_argStr.StartsWith(completeParamExpression))
            {
                out_paramValue = in_argStr.Replace(completeParamExpression, "");
                return true;
            }
            else
            {
                return false;
            }
        }
    }
}
